/**
*	SmallBasic pseudo-compiler.	Converts the source to byte-code.
*
*	2000-05-27, Nicholas Christopoulos
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

#define SCAN_MODULE
#include "sys.h"
#include "str.h"
#include "panic.h"
#include "device.h"
#include "kw.h"
#include "bc.h"
#include "smbas.h"

void	bc_scan_cmd(char *text) 				SEC(BCSCAN);
int		bc_scan_inline_if(char *text) 			SEC(BCSCAN);
int		bc_is_keyword(char *name) 				SEC(BCSCAN);
word	bc_search(word ip, byte code) 			SEC(BCSCAN);

#if defined(_PalmOS) 
static char	bc_fileName[65];
#else
static char	bc_fileName[1024];
#endif

#define	TEXT_LINE_SIZE		512

extern void	expr_parser(bc_t *bc)			 	SEC(BCSCAN);

struct keyword_s {
	char	name[16];		/* keyword */
	byte	code;			/* byte-code */
	};

struct keyword_s keyword_table[] = {
{ "$$$_int", 	kwTYPE_INT		},
{ "$$$_dbl",	kwTYPE_NUM		},
{ "$$$_str",	kwTYPE_STR		},
{ "$$$_log",	kwTYPE_LOGOPR	},
{ "$$$_cmp",	kwTYPE_CMPOPR	},
{ "$$$_add",	kwTYPE_ADDOPR	},
{ "$$$_mul",	kwTYPE_MULOPR	},
{ "$$$_pow",	kwTYPE_POWOPR	},
{ "$$$_unr",	kwTYPE_UNROPR	},
{ "$$$_var",	kwTYPE_VAR	},
{ "$$$_line",	kwTYPE_LINE	},
{ "$$$_LPAR",	kwTYPE_LEVEL_BEGIN	},
{ "$$$_RPAR",	kwTYPE_LEVEL_END	},
{ "$$$_CRVAR",	kwTYPE_CRVAR	},
{ "LOCAL", 		kwLOCAL	},
{ "SUB", 		kwPROC	},
{ "FUNC", 		kwFUNC	},
{ "BYREF",		kwBYREF	},
{ "DECLARE",	kwDECLARE	},

{ "LET", 		kwLET	},
{ "CONST", 		kwCONST	},
{ "DIM", 		kwDIM	},
{ "STOP",		kwSTOP	},
{ "END",		kwEND	},
{ "PRINT",		kwPRINT	},
{ "INPUT",		kwINPUT	},
{ "PEN",		kwPEN	},
{ "CLS",		kwCLS	},
{ "REM",		kwREM	},
{ "CHAIN",		kwCHAIN	},

{ "ON",			kwON	},
{ "OFF",		kwOFF	},

{ "LABEL",		kwLABEL	},
{ "GOTO",		kwGOTO	},
{ "IF",			kwIF	},
{ "THEN",		kwTHEN	},
{ "ELSE",		kwELSE	},
{ "ELIF",		kwELIF	},
{ "ELSEIF",		kwELIF	},
{ "ENDIF",		kwENDIF	},
{ "FI",			kwENDIF	},
{ "FOR",		kwFOR	},
{ "TO",			kwTO	},
{ "STEP",		kwSTEP	},
{ "NEXT",		kwNEXT	},
{ "WHILE",		kwWHILE	},
{ "WEND",		kwWEND	},
{ "REPEAT",		kwREPEAT	},
{ "UNTIL",		kwUNTIL	},
{ "GOSUB",		kwGOSUB	},
{ "RETURN",		kwRETURN	},
{ "READ",		kwREAD	},
{ "DATA",		kwDATA	},
{ "RESTORE",	kwRESTORE	},
{ "EXIT",		kwEXIT	},
{ "LOOP",		kwLOOP	},

{ "RND", 		kwRND	},
{ "ABS",		kwABS	},
{ "LEN",		kwLEN	},
{ "COS",		kwCOS	},
{ "SIN",		kwSIN	},
{ "TAN",		kwTAN	},
{ "ACOS",		kwACOS	},
{ "ASIN",		kwASIN	},
{ "ATAN",		kwATAN	},
{ "ATN",		kwATAN	},
{ "ACOSH",		kwACOSH	},
{ "ASINH",		kwASINH	},
{ "ATANH", 		kwATANH	},
{ "ATNH",		kwATANH	},
{ "COSH",		kwCOSH	},
{ "SINH",		kwSINH	},
{ "TANH",		kwTANH	},
{ "EXP",		kwEXP	},
{ "LOG",		kwLOG	},
{ "LOG10",		kwLOG10	},
{ "SQR",		kwSQR	},
{ "INT",		kwINT	},
{ "CINT",		kwCINT	},
{ "CDBL",		kwCDBL	},
{ "FIX",		kwFIX	},
{ "SGN",		kwSGN	},
{ "ASC", 		kwASC	},
{ "VAL", 		kwVAL	},
{ "RANDOMIZE",	kwRANDOMIZE	},
{ "ATAN2",		kwATAN2	},
{ "POW",		kwPOW	},
{ "ROUND",		kwROUND	},
{ "DEG",		kwDEG	},
{ "RAD",		kwRAD	},
{ "LBOUND",		kwLBOUND	},
{ "UBOUND",		kwUBOUND	},
{ "ERASE", 		kwERASE	},

{ "CHR",		kwCHR	},
{ "HEX",		kwHEX	},
{ "OCT",		kwOCT	},
{ "LCASE",		kwLCASE	},
{ "UCASE",		kwUCASE	},
{ "LTRIM",		kwLTRIM	},
{ "RTRIM",		kwRTRIM	},
{ "STRING"	,	kwSTRING	},
{ "STR"	,		kwSTR	},
{ "LEFT",		kwLEFT	},
{ "RIGHT",		kwRIGHT	},
{ "INSTR",		kwINSTR	},
{ "MID",		kwMID	},
{ "SPACE",		kwSPACE	},
{ "TIME",		kwTIME	},
{ "DATE",		kwDATE	},
{ "SPLIT",		kwWSPLIT	},

{ "AT",			kwAT	},
{ "LOCATE",		kwLOCATE	},

{ "TICKSPERSEC", kwTICKSPERSEC	},
{ "TICKS",		kwTICKS	},
{ "TIMER",		kwTIMER	},
{ "FRE",		kwFRE	},

{ "PSET",		kwPSET	},
{ "LINE",		kwLINE	},
{ "RECT",		kwRECT	},
{ "COLOR",		kwCOLOR	},
{ "FILLED",		kwFILLED	},
{ "CIRCLE",		kwCIRCLE	},

{ "BEEP",		kwBEEP	},
{ "SOUND",		kwSOUND	},
{ "PAUSE",		kwPAUSE	},

{ "INKEY",		kwINKEY	},

{ "TAB",		kwTAB	},
{ "CAT",		kwCAT	},
{ "DELAY",		kwDELAY	},
{ "ARC",		kwARC	},
{ "DRAW",		kwDRAW	},
{ "PLAY",		kwPLAY	},
{ "RUN",		kwRUN	},
{ "TXTW",		kwTEXTWIDTH	},
{ "TXTH",		kwTEXTHEIGHT	},
{ "CHART",		kwCHART	},
{ "MAX",		kwMAX	},
{ "MIN",		kwMIN	},
{ "DRAWPOLY",	kwDRAWPOLY	},

/* file I/O */
{ "FREEFILE",	kwFREEFILE	},
{ "OPEN", 		kwOPEN	},
{ "OUTPUT",		kwOUTPUT	},	// OPEN's args
{ "APPEND",		kwAPPEND	},	// OPEN's args
{ "AS",			kwAS },		// OPEN's args
{ "CLOSE", 		kwCLOSE  },
{ "LINEINPUT", 	kwLINEINPUT },		// The QB's keyword is "LINE INPUT"
{ "EOF", 		kwEOF	 },
{ "KILL", 		kwKILL	 },
{ "EXIST", 		kwEXIST	},
{ "SEEK", 		kwSEEK	},
{ "LOF", 		kwLOF	},
{ "COPY", 		kwCOPY	},
{ "RENAME", 	kwRENAME	},
//{ "INPUT", 		kwFINPUT },	// not needed

/* DEBUG */
{ "TRON",		kwTRON	},
{ "TROFF",		kwTROFF	},
{ "LOGPRINT",	kwLOGPRINT	},
#ifdef BC_DEBUG
{ "BCDUMP",		kwBCDUMP	},
{ "STKDUMP",	kwSTKDUMP	},
#endif

{ "", 0 }
};

/*
*	Notes:
*		block_level = the depth of nested block
*		block_id	= unique number of each block (based on stack use)
*
*		Example:
*		? xxx			' level 0, id 0
*		for i=1 to 20	' level 1, id 1
*			? yyy		' level 1, id 1
*			if a=1 		' level 2, id 2 (our IF uses stack)
*				...		' level 2, id 2
*			else		' level 2, id 2	// not 3
*				...		' level 2, id 2
*			fi			' level 2, id 2
*			if a=2		' level 2, id 3
*				...		' level 2, id 3
*			fi			' level 2, id 3
*			? zzz		' level 1, id 1
*		next			' level 1, id 1
*		? ooo			' level 0, id 0
*/

int			scan_error;
int			scan_line;			// source line count
static int	block_level;		// block level (FOR-NEXT,IF-FI,etc)
static int	block_id;			// unique ID for blocks (FOR-NEXT,IF-FI,etc)
static word	first_data_ip = 0xFFFF;

// ndc: 2001-03-01 static allocation (non tmp_alloc()) does not use the dynamic RAM and reduces the fragmentation
static char	*bc_name;
static char	*bc_parm;
static char	*bc_temp;
static char *bc_proc;
static char	bc_sec[33];
static int  bc_proc_level = 0;

static bc_t	bc_prog;
static bc_t	bc_data;

#define	GROWSIZE	128

/*
*	variable node
*/
struct var_s	{
	char	*name;
	};
typedef struct var_s var_t;

static var_t	*var_table;
static word		var_count;
static word		var_size;

/*
*	LABELS LIST
*/
struct label_s	{
//	char	*name;
	char	name[33];
	word	ip;			// instruction pointer
	byte	level;		// block level (used for GOTOs)
	word	block_id;	// block_id (FOR-NEXT,IF-FI,etc) used for GOTOs
	word	dp;			// data pointer
	};
typedef struct label_s label_t;

//static label_t	*lab_table;
//static word		lab_count;
//static word		lab_size;
static dbt_t	lab_table;
static word		lab_count;

/*
*	PROCEDURES/FUNCTIONS LIST
*/
struct proc_s	{
	char	*name;
	word	ip;			// instruction pointer
	word	vid;		// variable index (for functions)
	byte	level;		// block level (used for GOTOs)
	word	block_id;	// block_id (FOR-NEXT,IF-FI,etc) used for GOTOs
	};
typedef struct proc_s proc_t;

static proc_t	*proc_table;
static word		proc_count;
static word		proc_size;

/*
*	PASS2 - STACK
*/
struct pass_node_s	{
	char	sec[33];
	word	pos;		// instruction position
	word	line;		// the source-text line
	byte	level;		// block level
	word	block_id;	// block ID	
	};
typedef struct pass_node_s pass_node_t;

static dbt_t	bc_stack;
static word		stk_count;

/**
*	raise a compiler error
*/
void	sc_raise(const char *fmt, ...)
{
	char	*buff;
	va_list ap;

	va_start(ap, fmt);

	scan_error = 1;

	buff = tmp_alloc(TEXT_LINE_SIZE);
	#if defined(_PalmOS)
	StrVPrintF(buff, fmt, ap);
	#else
	vsprintf(buff, fmt, ap);
	#endif
	va_end(ap);
	
	dev_printf("\n* ERROR AT %s:%d *\n> %s\n", bc_sec, scan_line, buff);

	tmp_free(buff);
}

/*
*/
char*	bc_prepname(char *dest, const char *source, int size) SEC(BCSCAN);
char*	bc_prepname(char *dest, const char *source, int size)
{
	char	*p = (char *) source;

	while ( *p == ' ' )		p ++;
	strncpy(dest, p, size);
	dest[size] = '\0';
	p = dest;
	while ( *p && (is_alpha(*p) || is_digit(*p) || *p == '$' || *p == '/') )
		p ++;
	*p = '\0';

	str_alltrim(dest);
	return dest;
}

/*
*	returns the ID of the label. If there is no one, then it creates one
*/
int		bc_get_label_id(const char *label_name) SEC(BCSCAN);
int		bc_get_label_id(const char *label_name)
{
	int		idx = -1;
	int		i;
	char	name[33];
	label_t	label;

	bc_prepname(name, label_name, 32);

	for ( i = 0; i < lab_count; i ++ )	{
		dbt_read(lab_table, i, &label, sizeof(label_t));
		if	( strcmp(label.name, name) == 0 )	{
			idx = i;
			break;
			}
		}

	if	( idx == -1 )	{
/*
		if ( lab_count >= lab_size ) {
			lab_size += GROWSIZE;
			lab_table = (label_t *) tmp_realloc(lab_table, lab_size * sizeof(label_t));
			}

		idx = lab_count;
		lab_count ++;

		lab_table[idx].name = tmp_alloc(strlen(name)+1);
		strcpy(lab_table[idx].name, name);
		lab_table[idx].ip = 0xFFFF;
		lab_table[idx].dp = 0xFFFF;
		lab_table[idx].level = block_level;
		lab_table[idx].block_id = block_id;
*/
		strcpy(label.name, name);
		label.ip = 0xFFFF;
		label.dp = 0xFFFF;
		label.level = block_level;
		label.block_id = block_id;

		dbt_write(lab_table, lab_count, &label, sizeof(label_t));
		idx = lab_count;
		lab_count ++;
		}

	return idx;
}

/*
*	set LABEL's position (IP)
*/
void	bc_set_label_ip(int idx) SEC(BCSCAN);
void	bc_set_label_ip(int idx)
{
	label_t	label;

	dbt_read(lab_table, idx, &label, sizeof(label_t));
/*	lab_table[idx].ip = bc_prog.count;
	lab_table[idx].dp = bc_data.count;
	lab_table[idx].level = block_level;
	lab_table[idx].block_id = block_id;
*/
	label.ip = bc_prog.count;
	label.dp = bc_data.count;
	label.level = block_level;
	label.block_id = block_id;
	dbt_write(lab_table, idx, &label, sizeof(label_t));
}

/*
*/
int		bc_get_proc_id(const char *proc_name) SEC(BCSCAN);
int		bc_get_proc_id(const char *proc_name)
{
	int		idx = -1;
	int		i;
	char	tmp[33];
	char	*name = bc_temp;

	if	( bc_proc_level )	{
		// search the procedure's table for this LOCAL PROC
		bc_prepname(tmp, proc_name, 32);
		strcpy(name, bc_proc);
		strcat(name, "/");
		strcat(name, tmp);
		for ( i = 0; i < proc_count; i ++ )	{
			if	( strcmp(proc_table[i].name, name) == 0 )	{
				idx = i;
				break;
				}
			}
		}

	if	( idx == -1 )	{
		// search the procedure's table for this GLOBAL PROC
		bc_prepname(name, proc_name, 32);
		for ( i = 0; i < proc_count; i ++ )	{
			if	( strcmp(proc_table[i].name, name) == 0 )	{
				idx = i;
				break;
				}
			}
		}

	return idx;
}

/*
*/
int		bc_addproc(const char *proc_name) SEC(BCSCAN);
int		bc_addproc(const char *proc_name)
{
	char	tmp[33];
	char	*name = bc_temp;
	int		idx = -1, i;

	//
	bc_prepname(tmp, proc_name, 32);
	if	( bc_proc_level )	{
		strcpy(name, bc_proc);
		strcat(name, "/");
		strcat(name, tmp);
		}
	else
		strcpy(name, tmp);

	for ( i = 0; i < proc_count; i ++ )	{
		if	( strcmp(proc_table[i].name, name) == 0 )	{
			idx = i;
			break;
			}
		}

	if	( idx == -1 )	{
		if ( proc_count >= proc_size ) {
			proc_size += GROWSIZE;
			proc_table = tmp_realloc(proc_table, proc_size * sizeof(proc_t));
			}

		if	( !is_alpha(proc_name[0]) )
			sc_raise("WRONG PROC NAME: %s", proc_name);
		else	{
			proc_table[proc_count].name = tmp_alloc(strlen(name)+1);
			proc_table[proc_count].ip = 0xFFFF; //bc_prog.count;
			proc_table[proc_count].level = block_level;
			proc_table[proc_count].block_id = block_id;
			strcpy(proc_table[proc_count].name, name);
			idx = proc_count;
			proc_count ++;
			}
		}

	return idx;
}

int		bc_proc_setip(const char *proc_name, word ip) SEC(BCSCAN);
int		bc_proc_setip(const char *proc_name, word ip)
{
	int		idx;
	
	idx = bc_get_proc_id(proc_name);
	if	( idx != -1 )	{
		proc_table[idx].ip = bc_prog.count;
		proc_table[idx].level = block_level;
		proc_table[idx].block_id = block_id;
		}
	return idx;
}

word	bc_proc_getip(const char *proc_name) SEC(BCSCAN);
word	bc_proc_getip(const char *proc_name)
{
	int		idx;
	
	idx = bc_get_proc_id(proc_name);
	if	( idx != -1 )	
		return proc_table[idx].ip;
	return 0xFFFF;
}

/*
*/
char	*get_param_sect(char *text, const char *delim, char *dest) SEC(BCSCAN);
char	*get_param_sect(char *text, const char *delim, char *dest)
{
	char	*p = (char *) text;
	char	*d = dest;
	int		quotes = 0, level = 0, skip_ch = 0;

	if	( p == NULL )	{
		*dest = '\0';
		return 0;
		}

	while ( is_space(*p) )	p ++;

	while ( *p )	{
		if	( quotes )	{
			if	( *p == '\"' )
				quotes = 0;
			}
		else	{
			switch ( *p )	{
			case	'\"':
				quotes = 1;
				break;
			case	'(':
				level ++;
				break;
			case	')':
				level --;
				break;
			case	'\n':
			case	'\r':
				skip_ch = 1;
				break;
				};
			}

		// delim check
		if	( delim != NULL && level <= 0 && quotes == 0 )	{
			if	( strchr(delim, *p) != NULL )	
				break;
			}

		// copy
		if	( !skip_ch )	{
			*d = *p;
			d ++;
			}
		else
			skip_ch = 0;

		p ++;
		}

	*d = '\0';
	
	if	( quotes )
		sc_raise("MISSING (\")");
	if	( level > 0 )
		sc_raise("MISSING \')\'");
	if	( level < 0 )
		sc_raise("MISSING \'(\'");

	str_alltrim(dest);
	return p;
}

/*
*/
int		bc_get_error(void) SEC(BCSCAN);
int		bc_get_error()
{
	return scan_error;
}

/*
*	checking for missing labels
*/
void	bc_check_labels(void) SEC(BCSCAN);
void	bc_check_labels()
{
	int		i;
	label_t	label;

	for ( i = 0; i < lab_count; i ++ )	{
		dbt_read(lab_table, i, &label, sizeof(label_t));
		if	( label.ip == 0xFFFF )	{
			sc_raise("LABEL NOT DEFINED: %s", label.name);
			break;
			}
		}
}

/*
*	returns the id of the variable 'name'
*/
int		bc_get_var_id(const char *var_name) SEC(BCSCAN);
int		bc_get_var_id(const char *var_name)
{
	int		idx = -1;
	int		i;
	char	tmp[33];
	char	*name = bc_temp;

	if	( bc_proc_level )	{
		// search local name-space
		bc_prepname(tmp, var_name, 32);
		#if defined(_PalmOS)
		StrPrintF(name, "%s/%s", bc_proc, tmp);
		#else
		sprintf(name, "%s/%s", bc_proc, tmp);
		#endif
		for ( i = 0; i < var_count; i ++ )	{
			if	( strcmp(var_table[i].name, name) == 0 )	{
				idx = i;
				break;
				}
			}
		}

	if	( idx == -1 )	{
		// search global name-space
		bc_prepname(name, var_name, 32);
		for ( i = 0; i < var_count; i ++ )	{
			if	( strcmp(var_table[i].name, name) == 0 )	{
				idx = i;
				break;
				}
			}
		}

	if	( idx == -1 )	{
		// create one
		if ( var_count >= var_size ) {
			var_size += GROWSIZE;
			var_table = tmp_realloc(var_table, var_size * sizeof(var_t));
			}

		if	( !is_alpha(var_name[0]) )
			sc_raise("WRONG VAR NAME: %s", var_name);
		else	{
			var_table[var_count].name = tmp_alloc(strlen(name)+1);
			strcpy(var_table[var_count].name, name);
			idx = var_count;
			var_count ++;
			}
		}

	return idx;
}

/*
*	adds a mark in stack at the current code position
*/
void	bc_push1(word ip) SEC(BCSCAN);
void	bc_push1(word ip)
{
/*
	if ( stk_count >= stk_size ) {
		stk_size += GROWSIZE;
		bc_stack = tmp_realloc(bc_stack, stk_size * sizeof(pass_node_t));
		}

	bc_stack[stk_count].pos      = ip;
	bc_stack[stk_count].level    = block_level;
	bc_stack[stk_count].block_id = block_id;
	bc_stack[stk_count].line     = scan_line;
	bc_stack[stk_count].sec      = tmp_alloc(strlen(bc_sec)+1);
	strcpy(bc_stack[stk_count].sec, bc_sec);
	stk_count ++;
*/
	pass_node_t		node;

	strcpy(node.sec, bc_sec);
	node.pos      = ip;
	node.level    = block_level;
	node.block_id = block_id;
	node.line     = scan_line;
	dbt_write(bc_stack, stk_count, &node, sizeof(pass_node_t));

	stk_count ++;
}

/*
*	adds a mark in stack at the current code position
*/
void	bc_push(void) SEC(BCSCAN);
void	bc_push()
{
	bc_push1(bc_prog.count);
}

/*
*	returns the keyword code 
*/
int		bc_is_keyword(char *name)
{
	int		i, idx;
	byte	dolar_bug = 0;

//	Code to enable the $ but not for keywords (INKEY$=INKEY, PRINT$=PRINT !!!)
//	I don't want to increase the size of keywords table.
	idx = strlen(name) - 1;
	if	( name[idx] == '$' )	{
		name[idx] = '\0';
		dolar_bug ++;
		}

	for ( i = 0; keyword_table[i].name[0] != '\0'; i ++ )	{
		if	( strcmp(keyword_table[i].name, name) == 0 )
			return keyword_table[i].code;
		}

	if	( dolar_bug )
		name[idx] = '$';

	return -1;
}

/*
*	get keyword name
*/
char	*bc_getkeywordstr(char *text, char *dest) SEC(BCSCAN);
char	*bc_getkeywordstr(char *text, char *dest)
{
	char	*p = (char *) text;
	char	*d = dest;

	if	( p == NULL )	{
		*dest = '\0';
		return 0;
		}

	while ( is_space(*p) )	p ++;

	if	( *p == '?' )	{
		strcpy(dest, "PRINT");
		p ++;
		while ( is_space(*p) )	p ++;
		return p;
		}

	if	( *p == '\'' || *p == '#' )	{
		strcpy(dest, "REM");
		p ++;
		while ( is_space(*p) )	p ++;
		return p;
		}

	while ( is_alnum(*p) )	{
		*d = *p;
		d ++; p ++;
		}

//	Code to kill the $
//	if	( *p == '$' )	
//		p ++;

//	Code to enable the $
	if	( *p == '$' )
		*d ++ = *p ++;

	*d = '\0';
	while ( is_space(*p) )	p ++;
	return p;
}

/*
*	expression (parameters)
*/
void	bc_scan_expr(char *expr, byte no_parser) SEC(BCSCAN);
void	bc_scan_expr(char *expr, byte no_parser)
{
	char	*ptr = (char *) expr;
	int		idx, level = 0, check_udf = 0;
	int		tp;
	word	w, stip, cip;
	long	lv = 0;
	double	dv = 0;
	bc_t	bc;

	str_alltrim(expr);

	if	( *ptr == '\0' )	
		return;

	bc_create(&bc);

	while ( *ptr )	{

		if	( is_digit(*ptr) || *ptr == '.' || (*ptr == '&' && strchr("XHOB", *(ptr+1) ) ) )	{
			//	number - constant
			ptr = get_numexpr(ptr, bc_name, &tp, &lv, &dv);
			switch ( tp )	{
			case	1:
				bc_add2l(&bc, kwTYPE_INT, lv);
				continue;
			case	2:
				bc_add2d(&bc, kwTYPE_NUM, dv);
				continue;
			default:
				sc_raise("NUMERIC EXPR: SYNTAX ERROR");
				}
			}
		else if ( *ptr == '\'' /* || *ptr == '#' */ )  	// remarks
			break;		// WHILE
		else if	( is_alpha(*ptr) || *ptr == '?' )	{
			// keyword
			ptr = bc_getkeywordstr(ptr, bc_name);
			idx = bc_is_keyword(bc_name);
			if	( idx != -1 )	{
				switch ( idx )	{
				case	kwFOR:		// this FOR is separator
					bc_add1(&bc, kwFORSEP);
					break;
				case	kwINPUT:	// this INPUT is separator
					bc_add1(&bc, kwINPUTSEP);
					break;
				// invalid keywords
				case kwLABEL: case kwGOTO:	case kwGOSUB:	case kwRETURN:
				case kwIF:	  case kwWHILE:	case kwREPEAT:
				case kwELSE:  case kwELIF:  case kwENDIF:	case kwNEXT:	case kwWEND:
				case kwUNTIL: case kwPRINT: case kwLOCATE:
				case kwAT:	  case kwRECT:	case kwLINE:	case kwCLS:
				case kwCIRCLE: case kwLET:	case kwCONST:	case kwRANDOMIZE:
				case kwPSET:  case kwBEEP:	case kwSOUND:	case kwCHAIN:	case kwREAD: case kwDATA: case kwRESTORE:
				case kwDIM:	case kwARC:		case kwDRAW:	case kwPLAY:	case kwOPEN:
					sc_raise("%s: MUST BE FIRST STATEMENT ON THE LINE", bc_name);
					break;
				// null keywords
				case kwTHEN:
					if	( level )
						sc_raise("MISSING ')'");
					break;
				// valid keywords
				default:
					if	( idx == kwLBOUND || idx == kwUBOUND )	{	// LBOUND & UBOUND IS A SPECIAL CASE
						bc_add1(&bc, idx);

						while ( *ptr == ' ' ) ptr ++;
						if	( *ptr != '(' )
							sc_raise("U|LBOUND: MISSING '('");
						else	{
							level ++;
							bc_add1(&bc, kwTYPE_LEVEL_BEGIN);

							ptr ++;
							ptr = bc_getkeywordstr(ptr, bc_name);
	
							w = bc_get_var_id(bc_name);
							bc_add2i(&bc, kwTYPE_VAR, w);
							}
						}
					else	{
						bc_add1(&bc, idx);
						}
					}
				}
			else {	// idx == -1
				// keyword/operators or variable
				if	( strcmp(bc_name, "AND") == 0 )	{
					bc_add2(&bc, kwTYPE_LOGOPR, '&');
					}
				else if	( strcmp(bc_name, "OR") == 0 )	{ 
					bc_add2(&bc, kwTYPE_LOGOPR, '|');
					}
				else if	( strcmp(bc_name, "XOR") == 0 )	{ 
					bc_add2(&bc, kwTYPE_LOGOPR, '~');
					}
				else if	( strcmp(bc_name, "NOT") == 0 )	{ 
					bc_add2(&bc, kwTYPE_UNROPR, '!');
					}
				else if	( strcmp(bc_name, "MOD") == 0 )	{ 
					bc_add2(&bc, kwTYPE_MULOPR, '%');
					}
				else	{
					int		udf;

					udf = bc_get_proc_id(bc_name);
					if	( udf != -1 )	{
						// UDF

						bc_add2i(&bc, kwTYPE_CALL_UDF, udf);
						bc_add1i(&bc, 0);
						check_udf ++;
						}
					else	{
						// VARIABLE
						
						while ( *ptr == ' ' ) ptr ++;
						if	( *ptr == '(' )	{
							if	( *(ptr+1) == ')' )	{
								// null array
								ptr += 2;
								}
							}

						w = bc_get_var_id(bc_name);
						bc_add2i(&bc, kwTYPE_VAR, w);
						}
					}
				}
			}
		else if ( *ptr == ',' || *ptr == ';' || *ptr == '#' )	{
			// parameter separator
			bc_add2(&bc, kwTYPE_SEP, *ptr);

			ptr ++;
			}
		else if ( *ptr == '\"' )	{
			// string
			ptr = bc_store_string(&bc, ptr);
			}
		else if ( *ptr == '(' )	{
			// parenthesis
			level ++;
			bc_add1(&bc, kwTYPE_LEVEL_BEGIN);

			ptr ++;
			}
		else if ( *ptr == ')' )	{
			// parenthesis
			bc_add1(&bc, kwTYPE_LEVEL_END);

			level --;
			ptr ++;
			}
		else if ( is_space(*ptr) )	
			// null characters
			ptr ++;
		else	{
			// operators
			if	( *ptr == '+' || *ptr == '-' )	{
				bc_add2(&bc, kwTYPE_ADDOPR, *ptr);
				}
			else if	( *ptr == '*' || *ptr == '/' || *ptr == '\\' || *ptr == '%' )	{
				bc_add2(&bc, kwTYPE_MULOPR, *ptr);
				}
			else if	( *ptr == '^' )	{
				bc_add2(&bc, kwTYPE_POWOPR, *ptr);
				}
			else if	( strncmp(ptr, "<=", 2) == 0 || strncmp(ptr, "=<", 2) == 0 )	{
				bc_add2(&bc, kwTYPE_CMPOPR, OPLOG_LE);

				ptr ++;
				}
			else if	( strncmp(ptr, ">=", 2) == 0 || strncmp(ptr, "=>", 2) == 0 )	{
				bc_add2(&bc, kwTYPE_CMPOPR, OPLOG_GE);

				ptr ++;
				}
			else if	( strncmp(ptr, "<>", 2) == 0 || strncmp(ptr, "!=", 2) == 0 )	{
				bc_add2(&bc, kwTYPE_CMPOPR, OPLOG_NE);

				ptr ++;
				}
			else if	( *ptr == '=' || *ptr == '>' || *ptr == '<' )	{
				bc_add2(&bc, kwTYPE_CMPOPR, *ptr);
				}
			else if	( *ptr == '&' || *ptr == '|' || *ptr == '~' )	{
				bc_add2(&bc, kwTYPE_LOGOPR, *ptr);
				}
			else if	( *ptr == '!' )	{
				bc_add2(&bc, kwTYPE_UNROPR, *ptr);
				}
			else
				sc_raise("UNKNOWN OPERATOR: '%c'", *ptr);

			ptr ++;
			}
		};

	if	( level )
		sc_raise("MISSING ')'");

	if	( !scan_error )	{
		if	( no_parser == 0 )	{
			// optimization
			bc_add1(&bc, kwTYPE_EOC);
			expr_parser(&bc);
			}

		if	( bc.count )	{
			stip = bc_prog.count;
			bc_append(&bc_prog, &bc);

			if	( check_udf )	{
				cip = stip;
				while ( (cip = bc_search(cip, kwTYPE_CALL_UDF)) != 0xFFFF )	{
					bc_push1(cip);
					cip += 5;
					}
				}
			}
		bc_eoc(&bc_prog);
		}

	// clean-up
	bc_destroy(&bc);
}

/*
*	Converts DATA commands to bytecode
*/
void	bc_scan_data(char *source) SEC(BCSCAN);
void	bc_scan_data(char *source)
{
	char	*ptr = source;
	char	*commap;
	long	lv = 0;
	double	dv = 0, sign = 1;
	char	*tmp = bc_temp;
	int		quotes;
	int		tp;

	while ( *ptr )	{
		while ( *ptr == ' ' )	ptr ++;

		if	( *ptr == '\0' )	
			break;
		else if	( *ptr == ',' )	{
			bc_add1(&bc_data, kwTYPE_EOC);
			ptr ++;
			}
		else	{
			// find the end of the element
			commap = ptr;
			quotes = 0;
			while ( *commap )	{
				if	( *commap == '\"' )
					quotes = !quotes;
				else if ( (*commap == ',') && (quotes == 0) )
					break;

				commap ++;
				}
			if	( *commap == '\0' )
				commap = NULL;

			if	( commap != NULL )
				*commap = '\0';

			if  ((*ptr == '-' || *ptr == '+') && strchr("0123356789.", *(ptr+1)))	{ 
				if	( *ptr == '-' )
					sign = -1;
				ptr ++;
				}
			else
				sign = 1;

			if	( is_digit(*ptr) || *ptr == '.' 
					|| (*ptr == '&' && strchr("XHOB", *(ptr+1))) )	{

				//	number - constant
				ptr = get_numexpr(ptr, tmp, &tp, &lv, &dv);
				switch ( tp )	{
				case	1:
					bc_add2l(&bc_data, kwTYPE_INT, lv * sign);
					break;
				case	2:
					bc_add2d(&bc_data, kwTYPE_NUM, dv * sign);
					break;
				default:
					sc_raise("NUMERIC EXPR: SYNTAX ERROR");
					}
				}
			else	{
				// add it as string
				if	( *ptr != '\"' )	{
					strcpy(tmp, "\"");
					strcat(tmp,  ptr);
					strcat(tmp, "\"");
					bc_store_string(&bc_data, tmp);
					if	( commap )
						ptr = commap;
					else
						ptr = ptr + strlen(ptr);
					}
				else
					ptr = bc_store_string(&bc_data, ptr);
				}

			if	( commap != NULL )	
				*commap = ',';
			}
		}

	bc_add1(&bc_data, kwTYPE_EOC);	// no bc_eoc
}

/*
*	Scans the 'source' for "names" separated by 'delims' and returns the elements (pointer in source)
*	into args array.
*
*	Returns the number of items
*/
int		bc_getlist(char *source, char_p_t *args, char *delims, int maxarg) SEC(BCSCAN);
int		bc_getlist(char *source, char_p_t *args, char *delims, int maxarg)
{
	char	*p, *ps;
	int		count = 0;

	ps = p = source;
	while ( *p )	{
		if	( strchr(delims, *p) ) {
			*p = '\0';
			args[count] = ps;
			count ++;
			if	( count == maxarg )
				return count;
			ps = p+1;
			}

		p ++;
		}

	if	( *ps )	{
		*p = '\0';
		args[count] = ps;
		count ++;
		}

	return count;
}

/*
*/
char*	bc_next_char(char *source)	SEC(BCSCAN);
char*	bc_next_char(char *source)
{
	char	*p = source;

	while ( *p )	{
		if	( *p != ' ' )
			return p;
		p ++;
		}
	return p;
}

/*
*/
int		bc_getlist_insep(char *source, char_p_t *args, char *sep, char *delims, int maxarg) SEC(BCSCAN);
int		bc_getlist_insep(char *source, char_p_t *args, char *sep, char *delims, int maxarg)
{
	char	*p = source;
	char	*ps;
	int		count = 0, level = 1;
	
	p = strchr(source, sep[0]);
	if	( p )	{
		ps = p+1;
		p ++;
		while ( *p )	{
			if	( *p == sep[1] )	{
				level --;
				if ( level == 0 )
					break;
				}
			else if	( *p == sep[0] )
				level ++;

			p ++;
			}

		if	( *p == sep[1] )	{
			*p = '\0';
			if	( strlen(ps) )	
				count = bc_getlist(ps, args, delims, maxarg);
			}
		else	
			sc_raise("MISSING '%c'", sep[1]);
		}
	return count;
}

/*
*	INLINE IFs
*
*	returns true if there is an inline IF
*
*	IF expr THEN ... ---> IF expr THEN (:) .... (:FI)
*	IF expr THEN ... ELSE ... ---> IF expr THEN (:) .... (:ELSE:) ... (:FI)
*/
int		bc_scan_inline_if(char *text)
{
	char	*p = (char *) text;	// *text points to 'expr'
	char	*pthen, *pelse;
	char	buf[TEXT_LINE_SIZE];

	if	( scan_error )		return 0;

	pthen = p;
	do	{
		pthen = str_istr(pthen+1, " THEN ");
		if	( pthen )	{
			// store the expression
			while ( *p == ' ' )	p ++;
			strcpy(buf, p);
			p = str_istr(buf, " THEN ");
			*p = '\0';

			// check for ':'
			p = pthen+6;
			while ( *p == ' ' )	p ++;

			if	( *p != ':' && *p != '\0' )	{
				// store the IF
				block_level ++;	block_id ++;
				bc_push();	bc_add1(&bc_prog,kwIF);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				bc_scan_expr(buf, 0);
				if	( scan_error )		return 0;
				// store EOC
				bc_add1(&bc_prog,kwTYPE_EOC);	//bc_eoc();

				// auto-goto 
				p = pthen + 6;
				while ( *p == ' ' )	p ++;
				if	( is_digit(*p) )	{
					// add goto
					strcpy(buf, "GOTO ");
					strcat(buf, p);
					}
				else
					strcpy(buf, p);

				// ELSE command
				// If there are more inline-ifs (nested) the ELSE belongs to the first IF (that's an error)
				pelse = str_istr(buf+1, "ELSE");
				if	( pelse )	{
					do	{
						if	( (*(pelse-1) == ' ' || *(pelse-1) == '\t') &&
							  (*(pelse+4) == ' ' || *(pelse+4) == '\t') )	{

							*pelse = '\0';

							// scan the commands before ELSE
							bc_scan_cmd(buf);
							// add EOC
							bc_eoc(&bc_prog);

							// auto-goto
							strcpy(buf, "ELSE:");
							p = pelse + 4;
							while ( *p == ' ' || *p == '\t' )	p ++;
							if	( is_digit(*p) )	{
								// add goto
								strcat(buf, "GOTO ");
								strcat(buf, p);
								}
							else
								strcat(buf, p);
							
							//
							break;
							}
						else
							pelse = str_istr(pelse+1, "ELSE");
						} while ( pelse != NULL );
					}

				// scan the rest commands
				bc_scan_cmd(buf);
				// add EOC
				bc_eoc(&bc_prog);

				// add ENDIF
				bc_push();	bc_add1(&bc_prog, kwENDIF);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				block_level --;

				return 1;
				}
			else	// *p == ':'
				return 0;
			}
		else
			break;

		} while ( pthen != NULL );

	return 0;	// false
}

/*
*	PASS1: scan source line
*/
void	bc_scan_cmd(char *text)
{
	char	*p;
	char	*lb_end;
	char	*last_cmd;
	int		idx, sharp, decl = 0, vattr;
	char	pname[33], vname[33];

	if	( scan_error )		return;

	str_alltrim(text);
	p = text;

	if	( *p == ':' )	{
		p ++;
		bc_scan_cmd(p);
		return;
		}

	// remark
	if	( *p == '\'' || *p == '#' )	return;

	// empty line
	if	( *p == '\0' )		return;

	lb_end = p = bc_getkeywordstr((char *)text, bc_name);
	last_cmd = p;
	p = get_param_sect(p, ":", bc_parm);

	/* check old style labels */
	if	( is_all_digits(bc_name) )	{
		str_alltrim(bc_name);
		idx = bc_get_label_id(bc_name);
		bc_set_label_ip(idx);
		if	( scan_error )		return;

		// continue
		last_cmd = p = bc_getkeywordstr(lb_end, bc_name);
		if	( strlen(bc_name) == 0 )	{
			if	( !p )	return;
			if	( *p == '\0' ) return;
			}
		p = get_param_sect(p, ":", bc_parm);
		}

	/* what is this ? */
	idx = bc_is_keyword(bc_name);

	if	( idx == kwLET )	{	// old-style keyword LET
		char	*p;
		idx = -1;
		p = bc_getkeywordstr(bc_parm, bc_name);
		strcpy(bc_parm, p);
		}
	else if ( idx == kwDECLARE )	{ // declaration
		char	*p;

		decl = 1;
		p = bc_getkeywordstr(bc_parm, bc_name);
		idx = bc_is_keyword(bc_name);
		strcpy(bc_parm, p);
		if	( idx != kwPROC && idx != kwFUNC )	{
			sc_raise("USE DECLARE WITH SUB|FUNC KEYWORD");
			return;
			}
		}

	//////
	sharp = (bc_parm[0] == '#');

	if	( bc_parm[0] == '=' && (idx != -1) )	{	// can't check arrays
		sc_raise("%s: IS KEYWORD", bc_name);
		return;                             
		}
	else if	( (bc_parm[0] == '=' || bc_parm[0] == '(') && (idx == -1) )	{
		// add let command
		bc_add2i(&bc_prog, kwLET, bc_get_var_id(bc_name));

		if	( bc_parm[0] == '(' )	{
			char	*p = strchr(bc_parm, '=');
			char	*n;
			
			if	( !p )	
				sc_raise("MISSING '='");
			else	{
				if	( *bc_next_char(bc_parm+1) == ')' )	{
					bc_scan_expr(p, 0);
					}
				else	{
					// ARRAY (LEFT)
					*p = '\0';
					n = p;
					str_alltrim(bc_parm);
					bc_parm[0] = ' ';
					bc_parm[strlen(bc_parm)-1] = ' ';
	
					bc_add1(&bc_prog,kwTYPE_LEVEL_BEGIN);
					bc_scan_expr(bc_parm, 0);
					bc_prog.ptr[bc_prog.count-1] = kwTYPE_LEVEL_END;
	
					*n = '=';
					bc_scan_expr(n, 0);
					}
				}
			}
		else
			bc_scan_expr(bc_parm, 0);
		}
	else if ( /* idx == kwLET || */ idx == kwCONST )	{
		p = bc_getkeywordstr(bc_parm, bc_name);
		p = get_param_sect(p, ":", bc_parm);

		if	( bc_is_keyword(bc_name) != -1 )
			sc_raise("%s: IS KEYWORD", bc_name);
		else	{
			if	( !(bc_parm[0] == '=' || bc_parm[0] == '(') )	
				sc_raise("MISSING '='");
			else	{
				bc_add2i(&bc_prog,idx, bc_get_var_id(bc_name));
				bc_scan_expr(bc_parm, 0);
				}
			}
		}
	else {
		// add generic command
		if	( idx != -1 )	{
			if	( idx == kwLABEL )	{
				str_alltrim(bc_parm);
				idx = bc_get_label_id(bc_parm);
				bc_set_label_ip(idx);
				}
			else if	( idx == kwEXIT )	{
				bc_add1(&bc_prog, idx);
				str_alltrim(bc_parm);
				if	( strlen(bc_parm) )	{
					idx = bc_is_keyword(bc_parm);
					if	( idx == kwFOR || idx == kwLOOP || idx == kwPROC || idx == kwFUNC )	
						bc_add1(&bc_prog, idx);
					else
						sc_raise("USE EXIT [FOR|LOOP|SUB|FUNC]");
					}
				}
			else if	( idx == kwDECLARE )	{
				}
			else if	( idx == kwPROC || idx == kwFUNC )	{
				//
				//	PROCEDURES/FUNCTIONS
				//
				char_p_t pars[64];
				char	*lpar_ptr;
				int		i, count, pidx;

				lpar_ptr = strchr(bc_parm, '(');
				if	( lpar_ptr )
					*lpar_ptr = '\0';

				bc_prepname(pname, bc_parm, 32);
				if	( decl )	{
					// declaration
					if	( bc_proc_getip(pname) == 0xFFFF )
						bc_addproc(pname);
					}
				else	{
					if	( bc_proc_getip(pname) != 0xFFFF )
						sc_raise("PROC/FUNC ALREADY DEFINED: %s", pname);
					else {

						if	( (pidx = bc_proc_setip(pname, bc_prog.count)) == -1 )	{
							pidx = bc_addproc(pname);
							bc_proc_setip(pname, bc_prog.count);
							}

						//
						bc_add1(&bc_prog, kwGOTO);
						bc_add1i(&bc_prog,0);
						bc_add1(&bc_prog,0);

						block_level ++; block_id ++;
						bc_push();
						bc_add1(&bc_prog, idx);

						//
						if	( bc_proc_level )	{
							strcat(bc_proc, "/");
							strcat(bc_proc, pname);
							}
						else
							strcpy(bc_proc, pname);

						if	( !scan_error )	{
							bc_proc_level ++;

							if	( idx == kwFUNC )	{	// setup dvar. {F}/{F}
								char	*buf;
								buf = tmp_alloc(strlen(bc_proc)+strlen(pname)+2);
								#if defined(_PalmOS)
								StrPrintF(buf, "%s/%s", bc_proc, pname);
								#else
								sprintf(buf, "%s/%s", bc_proc, pname);
								#endif
								proc_table[pidx].vid = bc_get_var_id(buf);
								tmp_free(buf);
								}
							else
								proc_table[pidx].vid = 0xFFFF;

							// parameters
							if	( lpar_ptr )	{
								char	*buf;

								buf = tmp_alloc(TEXT_LINE_SIZE);

								*lpar_ptr = '(';
								count = bc_getlist_insep(bc_parm, pars, "()", ",", 64);
								bc_add2c(&bc_prog, kwTYPE_PARAM, count);
								for ( i = 0; i < count; i ++ )	{
									if	( strncmp(pars[i], "BYREF ", 6) == 0 )	{
										bc_prepname(vname, pars[i]+6, 32);
										vattr = 0x80;
										}
									else	{
										bc_prepname(vname, pars[i], 32);
										vattr = 0;
										}
										
									if	( strchr(pars[i], '(') )	{
										vattr |= 1;
										}

									bc_add1(&bc_prog, vattr);
									#if defined(_PalmOS)
									StrPrintF(buf, "%s/%s", bc_proc, vname);
									#else
									sprintf(buf, "%s/%s", bc_proc, vname);
									#endif
									bc_add1i(&bc_prog, bc_get_var_id(buf));
									}

								tmp_free(buf);
								}
							else	{
								bc_add2c(&bc_prog, kwTYPE_PARAM, 0);
								}
							}
						}
					}
				}
			else if	( idx == kwLOCAL )	{	// local variables
				char_p_t pars[64];
				int		i, count;

				count = bc_getlist(bc_parm, pars, ",", 64);
				bc_add2c(&bc_prog, kwTYPE_CRVAR, count);
				for ( i = 0; i < count; i ++ )	{
					bc_prepname(vname, pars[i], 32);

					if	( bc_proc_level )	{
						char	*buf;
						buf = tmp_alloc(strlen(bc_proc)+strlen(vname)+2);
						#if defined(_PalmOS)
						StrPrintF(buf, "%s/%s", bc_proc, vname);
						#else
						sprintf(buf, "%s/%s", bc_proc, vname);
						#endif
						bc_add1i(&bc_prog, bc_get_var_id(buf));
						tmp_free(buf);
						}
					else
						bc_add1i(&bc_prog, bc_get_var_id(vname));
					}
				}
			else if ( idx == kwREM )	
				return;
			else if ( idx == kwGOTO )	{
				str_alltrim(bc_parm);
				bc_push();
				bc_add2i(&bc_prog, idx, bc_get_label_id(bc_parm));
				bc_add1(&bc_prog, block_level);
				}
			else if ( idx == kwGOSUB )	{
				str_alltrim(bc_parm);
				bc_add2i(&bc_prog,idx, bc_get_label_id(bc_parm));
				}
			//
			//	IF
			//
			else if ( idx == kwIF )	{
				// from here, we can scan for inline IF
				if	( bc_scan_inline_if(last_cmd) )	{
					// inline-IFs
					return;
					}
				else	{
					block_level ++; block_id ++;
					bc_push();	bc_add1(&bc_prog,idx);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
					bc_scan_expr(bc_parm, 0);
					bc_add1(&bc_prog,kwTYPE_EOC);	//bc_eoc();
					}
				}
			//
			//	ON x GOTO|GOSUB ...
			//
			else if ( idx == kwON )	{
				char	*p;
				int		keep_ip, count;
				char_p_t pars[64];
				int		i;
			
				idx = kwONJMP;		// WARNING!

				bc_push();
				bc_add1(&bc_prog,idx);
				bc_add1i(&bc_prog,0); bc_add1i(&bc_prog,0);	// true/false ip (pass2)

				if	( (p = strstr(bc_parm, " GOTO ")) != NULL )	{
					bc_add1(&bc_prog, kwGOTO);	// the command
					*p = '\0'; p += 6;
					keep_ip = bc_prog.count;
					bc_add1(&bc_prog, 0);			// the counter

					//count = bc_scan_label_list(p);
					count = bc_getlist(p, pars, ",", 64);
					for ( i = 0; i < count; i ++ )	
						bc_add1i(&bc_prog, bc_get_label_id(pars[i]));

					if	( count == 0 )
						sc_raise("ON x GOTO WHERE?");
					else	
						bc_prog.ptr[keep_ip] = count;

					bc_scan_expr(bc_parm, 0);	// the expression
					bc_eoc(&bc_prog);
					}
				else if	( (p = strstr(bc_parm, " GOSUB ")) != NULL )	{
					bc_add1(&bc_prog,kwGOSUB);	// the command
					*p = '\0';	p += 7;
					keep_ip = bc_prog.count;
					bc_add1(&bc_prog,0);			// the counter

					//count = bc_scan_label_list(p);
					count = bc_getlist(p, pars, ",", 64);
					for ( i = 0; i < count; i ++ )	
						bc_add1i(&bc_prog, bc_get_label_id(pars[i]));

					if	( count == 0 )
						sc_raise("ON x GOSUB WHERE?");
					else	
						bc_prog.ptr[keep_ip] = count;

					bc_scan_expr(bc_parm, 0);	// the expression
					bc_eoc(&bc_prog);
					}
				else
					sc_raise("ON WHAT?");
				}
			//
			//	FOR
			//
			else if ( idx == kwFOR )	{
				char	*p = strchr(bc_parm, '=');
				char	*p_lev;
				char	*n;
			
				block_level ++; block_id ++;
				bc_push();	
				bc_add1(&bc_prog, kwFOR);	
				bc_add1i(&bc_prog, 0);
				bc_add1i(&bc_prog, 0);

				if	( !p )	
					sc_raise("FOR: MISSING '='");
				else	{
					*p = '\0';
					n = p;

					strcpy(bc_name, bc_parm);
					str_alltrim(bc_name);
					if	( !is_alpha(*bc_name) )
						sc_raise("FOR: %s IS NOT A VARIABLE", bc_name);
					else	{
						p_lev = bc_name;
						while ( is_alnum(*p_lev) || *p_lev == ' ' )	p_lev ++;
						if	( *p_lev == '(' )	
							sc_raise("FOR: %s IS AN ARRAY", bc_name);
						else	{
							bc_add2i(&bc_prog, kwTYPE_VAR, bc_get_var_id(bc_name));

							*n = '=';
							bc_scan_expr(n+1, 0);
							}
						}
					}
				}
			//
			//	WHILE - REPEAT
			//
			else if ( idx == kwWHILE || idx == kwREPEAT )	{
				// WHILE & REPEAT DOES NOT USE STACK 
				block_level ++; block_id ++;
				bc_push();	bc_add1(&bc_prog,idx);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				bc_scan_expr(bc_parm, 0);
				}
			else if ( idx == kwELSE || idx == kwELIF )	{
				bc_push();	bc_add1(&bc_prog,idx);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				bc_scan_expr(bc_parm, 0);
				}
			else if ( idx == kwENDIF || idx == kwNEXT )	{
				bc_push();	bc_add1(&bc_prog,idx);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				block_level --;
				}
			else if ( idx == kwWEND || idx == kwUNTIL )	{
				bc_push();	bc_add1(&bc_prog,idx);	bc_add1i(&bc_prog,0);	bc_add1i(&bc_prog,0);
				// WHILE & REPEAT DOES NOT USE STACK
				block_level --;
				bc_scan_expr(bc_parm, 0);
				}
			else if (  idx == kwSTEP || idx == kwTO || idx == kwTHEN 
					|| idx == kwCOS  || idx == kwSIN || idx == kwLEN || idx == kwLOOP  )		// functions...
				{
				sc_raise("%s: WRONG POSITION", bc_name);
				}
			else if ( idx == kwRESTORE )	{
				bc_push();
				bc_add2i(&bc_prog, idx, bc_get_label_id(bc_parm));
				}
			else if ( idx == kwEND )	{
				if	( bc_proc_level )	{
					char	*dol;

					// UDP/F RETURN
					dol = strrchr(bc_proc, '/');
					if	( dol )
						*dol = '\0';
					else
						*bc_proc = '\0';

					bc_push();
					bc_add1(&bc_prog, kwTYPE_RET);

					bc_proc_level --;
					block_level --; block_id ++;
					}
				else	{
					// END OF PROG
					bc_add1(&bc_prog, idx);
					}
				}
			else if ( idx == kwSTOP )
				bc_add1(&bc_prog, idx);
			else if ( idx == kwDATA )	
				bc_scan_data(bc_parm);
			else if ( idx == kwINPUT && sharp )	{
				bc_add1(&bc_prog, kwFILEINPUT);
				bc_scan_expr(bc_parm, 0);
				}
			else if ( idx == kwPRINT && sharp )	{
				bc_add1(&bc_prog, kwFILEPRINT);
				bc_scan_expr(bc_parm, 0);
				}
			else	{
				bc_add1(&bc_prog, idx);
				bc_scan_expr(bc_parm, 0);
				}
			}
		else	{
			int		udp;

			udp = bc_get_proc_id(bc_name);
			if	( udp == -1 )
				sc_raise("%s: UNKNOWN COMMAND", bc_name);
			else	{
				bc_push();
				bc_add2i(&bc_prog, kwTYPE_CALL_UDP, udp);
				bc_add1i(&bc_prog, 0);
				bc_add1(&bc_prog, kwTYPE_LEVEL_BEGIN);
				bc_scan_expr(bc_parm, 0);
				bc_add1(&bc_prog, kwTYPE_LEVEL_END);
				}
			}
		}

	if	( *p == ':' )	{
		// command separator
		bc_eoc(&bc_prog);
		p ++;
		bc_scan_cmd(p);
		}
}

/*
*	skip command bytes
*/
word	bc_nextcmd(word ip) SEC(BCSCAN);
word	bc_nextcmd(word ip)
{
	word	code;
	word	len;

	code = bc_prog.ptr[ip];
	ip ++;

	switch ( code )	{
	case	kwEXIT:
		if	( bc_prog.ptr[ip] == kwFOR || bc_prog.ptr[ip] == kwLOOP ||
		      bc_prog.ptr[ip] == kwPROC || bc_prog.ptr[ip] == kwFUNC )
			ip ++;
		break;
	case	kwTYPE_SEP:
	case	kwTYPE_LOGOPR:
	case	kwTYPE_CMPOPR:
	case	kwTYPE_ADDOPR:
	case	kwTYPE_MULOPR:
	case	kwTYPE_POWOPR:
	case	kwTYPE_UNROPR:
		ip ++;
		break;
	case	kwLET:
	case	kwTYPE_LINE:
	case	kwLABEL:
	case	kwGOSUB:
	case	kwRESTORE:
	case	kwTYPE_VAR:
	case	kwTYPE_CALL_UDP:
	case	kwTYPE_CALL_UDF:
		ip += 2;
		break;
	case	kwGOTO:
		ip += 3;
		break;
	case	kwTYPE_CRVAR:
		len = bc_prog.ptr[ip];
		ip += (len * 2) + 1;
		break;
	case	kwTYPE_PARAM:
		len = bc_prog.ptr[ip];
		ip += (len * 3) + 1;
		break;
	case	kwONJMP:
		ip += 5;
		ip += (bc_prog.ptr[ip] << 1);
		break;
	case	kwTYPE_INT:
		ip += 4;
		break;
	case	kwTYPE_NUM:
		ip += 8;
		break;
	case	kwTYPE_STR:
		memcpy(&len, bc_prog.ptr+ip, 2);
		len += 2;
		ip += len;
		break;
	case	kwIF:		case	kwFOR: 		case	kwWHILE:	case	kwREPEAT:
	case	kwELSE:		case	kwELIF:
	case	kwENDIF:	case	kwNEXT:		case	kwWEND:		case	kwUNTIL:
		ip += 4;
		break;
		};
	return ip;
}

/*
*	search for command (in byte-code)
*/
word	bc_search(word ip, byte code)
{
	word	i = ip;

	do	{
		if	( code == bc_prog.ptr[i] )
			return i;

		i = bc_nextcmd(i);
		} while ( i < bc_prog.count );
	return 0xFFFF;
}

/*
*	search for End-Of-Command mark
*/
word	bc_search_eoc(word ip) SEC(BCSCAN);
word	bc_search_eoc(word ip)
{
	word	i = ip;
	word	code;

	do	{
		code = bc_prog.ptr[i];
		if	( code == kwTYPE_EOC || code == kwTYPE_LINE )
			return i;

		i = bc_nextcmd(i);
		} while ( i < bc_prog.count );
	return bc_prog.count;
}

/*
*	search stack
*/
word	bc_search_stack(word start, word code, word level) SEC(BCSCAN);
word	bc_search_stack(word start, word code, word level)
{
	word		i;
	pass_node_t	node;

	for ( i = start; i < stk_count; i ++ )	{
		dbt_read(bc_stack, i, &node, sizeof(pass_node_t));

		if	( node.level == level )	{
			if	( bc_prog.ptr[node.pos] == code )
				return node.pos;
			}
		}
	return 0xFFFF;
}

/*
*	search stack backward
*/
word	bc_search_stack_backward(word start, word code, word level) SEC(BCSCAN);
word	bc_search_stack_backward(word start, word code, word level)
{
	word		i = start;
	pass_node_t	node;

	for ( ; i < stk_count; i -- )	{	// WARNING: ITS UNSIGNED, SO WE'LL SEARCH IN RANGE [0..STK_COUNT]
		dbt_read(bc_stack, i, &node, sizeof(pass_node_t));
		if	( node.level == level )	{
			if	( bc_prog.ptr[node.pos] == code )
				return node.pos;
			}
		}
	return 0xFFFF;
}

/*
*	PASS 2 (write jumps for IF/FOR/WHILE/REPEAT)
*/
void	bc_pass2_scan(void) SEC(BCSCAN);
void	bc_pass2_scan()
{
	word	i, j, true_ip, false_ip, label_id, w;
	word	a_ip, b_ip, c_ip, count;
	byte	code, level;
	pass_node_t	node;
	label_t	label;

	dev_printf("\rPASS2: Node %d/%d", i, stk_count);

	for ( i = 0; i < stk_count; i ++ )	{

		if	( (i % 32) == 0 )
			dev_printf("\rPASS2: Node %d/%d", i, stk_count);

		dbt_read(bc_stack, i, &node, sizeof(pass_node_t));

		scan_line = node.line;
		strcpy(bc_sec, node.sec);

		code = bc_prog.ptr[node.pos];

		if	(  code != kwGOTO && code != kwRESTORE 
		    && code != kwTYPE_CALL_UDP && code != kwTYPE_CALL_UDF 
			&& code != kwPROC && code != kwFUNC && code != kwTYPE_RET )	{
			true_ip = bc_search_eoc(node.pos+5);
			memcpy(bc_prog.ptr+node.pos+1, &true_ip, 2);
			}

		switch ( code )	{
		case kwPROC:
		case kwFUNC:
			// update start's GOTO
			true_ip = bc_search_stack(i+1, kwTYPE_RET, node.level) + 1;
			memcpy(bc_prog.ptr+node.pos-3, &true_ip, 2);
			break;

		case kwRESTORE:
			memcpy(&label_id, bc_prog.ptr+node.pos+1, 2);
			dbt_read(lab_table, label_id, &label, sizeof(label_t));
			count = first_data_ip + label.dp;
			memcpy(bc_prog.ptr+node.pos+1, &count, 2);	// change LABEL-ID with DataPointer
			break;

		case kwTYPE_CALL_UDP:
		case kwTYPE_CALL_UDF:
			// update real IP
			memcpy(&label_id, bc_prog.ptr+node.pos+1, 2);
			true_ip = proc_table[label_id].ip + 5;
			memcpy(bc_prog.ptr+node.pos+1, &true_ip, 2);	

			// update return-var ID
			true_ip = proc_table[label_id].vid;
			memcpy(bc_prog.ptr+node.pos+3, &true_ip, 2);	
			break;

		case kwONJMP:
			// kwONJMP:1 trueip:2 falseip:2 command:1 count:1 label1:2 label2:2 ...
			count = bc_prog.ptr[node.pos+6];

			// change label IDs with the real IPs
			for ( j = 0; j < count; j ++ )	{
				memcpy(&label_id, bc_prog.ptr+node.pos+(j<<1)+7, 2);
				dbt_read(lab_table, label_id, &label, sizeof(label_t));
				w = label.ip;
				memcpy(bc_prog.ptr+node.pos+(j<<1)+7, &w, 2);
				}
			break;

		case kwGOTO:	// LONG JUMPS
			memcpy(&label_id, bc_prog.ptr+node.pos+1, 2);
			dbt_read(lab_table, label_id, &label, sizeof(label_t));
			w = label.ip;
			memcpy(bc_prog.ptr+node.pos+1, &w, 2);	// change LABEL-ID with IP
			level = bc_prog.ptr[node.pos+3];
			bc_prog.ptr[node.pos+3] = 0; // number of POPs ?????????

			if	( level >= label.level )
				bc_prog.ptr[node.pos+3] = level - label.level; // number of POPs
			else	
				bc_prog.ptr[node.pos+3] = 0; // number of POPs ?????????

			break;
		case kwFOR:
			a_ip = bc_search(node.pos+5, kwTO);
			false_ip = bc_search_stack(i+1, kwNEXT, node.level);

			if	( false_ip == 0xFFFF )	{
				sc_raise("'FOR' WITHOUT 'NEXT'");
				return;
				}
			if	( a_ip > false_ip || a_ip == 0xFFFF )	{
				sc_raise("'FOR' WITHOUT 'TO'");
				return;
				}

			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;
		case kwWHILE:
			false_ip = bc_search_stack(i+1, kwWEND, node.level);

			if	( false_ip == 0xFFFF )	{
				sc_raise("'WHILE' WITHOUT 'WEND'");
				return;
				}

			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;

		case kwREPEAT:
			false_ip = bc_search_stack(i+1, kwUNTIL, node.level);

			if	( false_ip == 0xFFFF )	{
				sc_raise("'REPEAT' WITHOUT 'UNTIL'");
				return;
				}

			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;

		case kwIF:
		case kwELIF:
			a_ip = bc_search_stack(i+1, kwELSE,  node.level);
			b_ip = bc_search_stack(i+1, kwENDIF, node.level);
			c_ip = bc_search_stack(i+1, kwELIF,  node.level);

			false_ip = a_ip;
			if	( b_ip < false_ip )
				false_ip = b_ip;
			if	( c_ip < false_ip )
				false_ip = c_ip;

			if	( false_ip == 0xFFFF )	{
				sc_raise("'IF' WITHOUT 'ENDIF'");
				return;
				}

			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;

		case kwELSE:
			false_ip = bc_search_stack(i+1, kwENDIF, node.level);

			if	( false_ip == 0xFFFF )	{
				sc_raise("'IF' WITHOUT 'ENDIF'");
				return;
				}

			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;

/////
/////
/////

		case kwTYPE_RET:
			break;

		case kwWEND:
			false_ip = bc_search_stack_backward(i-1, kwWHILE, node.level);
			if	( false_ip == 0xFFFF )	{
				sc_raise("'WEND' WITHOUT 'WHILE'");
				return;
				}
			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;
		case kwUNTIL:
			false_ip = bc_search_stack_backward(i-1, kwREPEAT, node.level);
			if	( false_ip == 0xFFFF )	{
				sc_raise("'UNTIL' WITHOUT 'REPEAT'");
				return;
				}
			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;
		case kwNEXT:
			false_ip = bc_search_stack_backward(i-1, kwFOR, node.level);
			if	( false_ip == 0xFFFF )	{
				sc_raise("'NEXT' WITHOUT 'FOR'");
				return;
				}
			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;
		case kwENDIF:
			false_ip = bc_search_stack_backward(i-1, kwIF, node.level);
			if	( false_ip == 0xFFFF )	{
				sc_raise("'ENDIF' WITHOUT 'IF'");
				return;
				}
			memcpy(bc_prog.ptr+node.pos+3, &false_ip, 2);
			break;
			};
		}

	dev_printf("\rPASS2: Node %d/%d\n", stk_count, stk_count);
}

/*
*	initialize
*/
void	bc_init() SEC(BCSCAN);
void	bc_init()
{
	#if defined(_PalmOS)
	dev_cls();
	#endif
	bc_name = tmp_alloc(TEXT_LINE_SIZE);
	bc_parm = tmp_alloc(TEXT_LINE_SIZE);
	bc_temp = tmp_alloc(TEXT_LINE_SIZE);
	bc_proc = tmp_alloc(TEXT_LINE_SIZE);

	scan_line = 0;
	scan_error = 0;
	lab_count = 0;
	var_count = 0;
	stk_count = 0;
	proc_count = 0;
	block_level = 0;
	block_id = 0;
	first_data_ip = 0xFFFF;
	bc_proc_level = 0;
	bc_proc[0] = '\0';

	var_table  = (var_t*)       tmp_alloc(GROWSIZE * sizeof(var_t));
	lab_table  = dbt_create("BCS-LBL.VMT");
	proc_table = (proc_t*)      tmp_alloc(GROWSIZE * sizeof(proc_t));
	bc_stack   = dbt_create("BCS-STK.VMT");

	var_size   = /* lab_size = stk_size = */ proc_size  = GROWSIZE;
	var_count  = lab_count = stk_count = proc_count = 0;

	bc_create(&bc_prog);
	bc_create(&bc_data);

	if	( !var_table || !lab_table || !proc_table || !bc_stack )	
		panic("bc_init(): OUT OF MEMORY");

	dev_printf("VMT Initialization...\n");
	dbt_prealloc(lab_table, os_cclabs1, sizeof(label_t));
	dbt_prealloc(bc_stack, os_ccpass2, sizeof(pass_node_t));

	/*
	*	create system variables
	*/
	bc_get_var_id("OSVER");
	bc_get_var_id("OSNAME");
	bc_get_var_id("SBVER");
	bc_get_var_id("PI");
	bc_get_var_id("XMAX");
	bc_get_var_id("YMAX");
	bc_get_var_id("BPP");
	bc_get_var_id("TRUE");
	bc_get_var_id("FALSE");
	bc_get_var_id("LINECHART");
	bc_get_var_id("BARCHART");
}

/*
*	clean up
*/
void	bc_close() SEC(BCSCAN);
void	bc_close()
{
	int		i;
	#if defined(_PalmOS)
//	LocalID		lid;
	#endif

	bc_destroy(&bc_prog);
	bc_destroy(&bc_data);

	for ( i = 0; i < var_count; i ++ )
		tmp_free(var_table[i].name);
	for ( i = 0; i < proc_count; i ++ )
		tmp_free(proc_table[i].name);

	tmp_free(var_table);
	dbt_close(lab_table);
	tmp_free(proc_table);
	dbt_close(bc_stack);
//	#if defined(_PalmOS)
//	if ( (lid = DmFindDatabase(0, "BCS-LBL.VMT")) != 0 )	DmDeleteDatabase(0, lid);
//	if ( (lid = DmFindDatabase(0, "BCS-STK.VMT")) != 0 )	DmDeleteDatabase(0, lid);
//	#endif

	var_count = lab_count = stk_count = proc_count = 0;

	tmp_free(bc_proc);
	tmp_free(bc_temp);
	tmp_free(bc_parm);
	tmp_free(bc_name);
}

/*
*	returns true if the 'fileName' exists
*/
int		basFileExist(const char *basfile)
{
	int		check = 0;
	char	*p, *fileName;
	#if defined(_PalmOS)
	LocalID	lid;
	#endif

	fileName = tmp_alloc(strlen(basfile)+1);
	strcpy(fileName, basfile);

	p = strchr(fileName, '.');
	if	( !p )	
		strcat(fileName, ".bas");
	else	{
		if	( str_istr(fileName, ".bas") == NULL )
			strcat(fileName, ".bas");
		}

	#if defined(_PalmOS)
	lid = DmFindDatabase(0, fileName);
	check = (lid != 0);
	#else
		#if defined(_Win32)
		check = (access(fileName, 0) == 0);
		#else
		check = (access(fileName, R_OK) == 0);
		#endif
	#endif

	tmp_free(fileName);
	return check;
}

/*
*/
void	bc_load(const char *fileName) SEC(BCSCAN);
void	bc_load(const char *fileName)
{
#if defined(_PalmOS)
	DmOpenRef	fp;
	LocalID		lid;
	int			l, i;
	VoidPtr		rec_p = NULL;
	VoidHand	rec_h = NULL;
#else
	int		unxh;
	struct stat st;
#endif

	strcpy(bc_fileName, fileName);
#if defined(_PalmOS)
	lid = DmFindDatabase(0, (char *) fileName);
	fp = DmOpenDatabase(0, lid, dmModeReadWrite);
	if	( !fp )	{
		panic("LOAD: CAN'T OPEN FILE %s", fileName);
		return;
		}
	l = DmNumRecords(fp) - 1;
	if	( l <= 0 )	{
		panic("LOAD: BAD FILE STRUCTURE %s", fileName);
		return;
		}

	for ( i = 0; i < l; i ++ )	{
		rec_h = DmGetRecord(fp, i+1);
		if	( !rec_h )
			panic("LOAD: CAN'T GET RECORD %s", fileName);
		rec_p = mem_lock(rec_h);
		if	( !rec_p )
			panic("LOAD: CAN'T LOCK RECORD %s", fileName);

		if	( i == 0 && strlen(rec_p+6) == 0 )
			bc_scan("Main", rec_p+70);	// + sizeof(sec_t);
		else
			bc_scan(rec_p+6, rec_p+70);	// + sizeof(sec_t);

		mem_unlock(rec_h);
		DmReleaseRecord(fp, i+1, 0);

		if	( bc_get_error() )
			break;
		}

	DmCloseDatabase(fp);
#else
	stat(fileName, &st);
	unxh = open(fileName, 0);
	if	( unxh == -1 )	{
		panic("can't open %s\n", fileName);
		return;
		}
	else	{
		char	*buff;

		/* read & compile file */
		buff = (char *) tmp_alloc(st.st_size+1);
		read(unxh, buff, st.st_size);
		buff[st.st_size] = '\0';

		bc_scan(NULL, buff);

		tmp_free(buff);
		close(unxh);
		}
#endif
}

/**
*	PASS 1
*/
void	bc_scan(const char *section, const char *text)
{
	char	*ps, *p, lc = 0;
	int		quotes, len;
	char	pname[33];
	char	*code_line;
	char	*new_text;

	code_line = tmp_alloc(TEXT_LINE_SIZE);
	new_text = tmp_alloc((len=(strlen(text) + 4)));
	memset(new_text, 0, len);

	/*
	*	Make it simple text
	*		* space-chars is only the space
	*		* CR/LF are fixed
	*		* control chars are out
	*/
	if	( section )
		strncpy(bc_sec, section, 32);
	else
		strncpy(bc_sec, "Main", 32);
	bc_sec[32] = '\0';

//	dev_printf("Preparing text...\n");
	p = (char *) text;
	ps = new_text;
	quotes = 0;
	while ( *p )	{
		if	( !quotes )	{
			if ( *p == '\n' )	{
				*ps ++ = '\n';
				while ( *(p+1) == ' ' || *(p+1) == '\t' )
					p ++;
				}
			else if ( (*p == '\t' || *p == ' ') && lc != ' ' )
				*ps ++ = ' ';
			else if	( (*p > ' ') || (*p & 0x80) )
				*ps ++ = to_upper(*p);

			lc = *ps;
			}
		else
			*ps ++ = *p;

		if	( *p == '\"' )	
			quotes = !quotes;

		p ++;
		}
	*ps = '\0';	
	strcat(new_text, "\n");

	/*
	*/
//	dev_printf("Scanning for routines...\n");
	p = ps = new_text;
	bc_proc_level = 0;
	*bc_proc = '\0';
	while ( *p )	{
		if	( strncmp("#INC:", p, 5) == 0 )	{
			char	*crp  = NULL;

			p += 5;
			if	( *p == '\"' )	{
				p ++;

				crp = p;
				while ( *crp != '\0' && *crp != '\"' )
					crp ++;

				if	( *crp == '\0' )	{
					sc_raise("#INC: MISSING \"");
					break;
					}

				*crp = '\0';
				lc = '\"';
				}
			else	{
				crp = strchr(p, '\n');
				*crp = '\0';
				lc = '\n';
				}

			strcpy(code_line, p);
			str_alltrim(code_line);
			if	( !basFileExist(code_line) )	
				sc_raise("FILE %s: FILE %s DOES NOT EXISTS", bc_fileName, code_line);
			else	{
				#if defined(_PalmOS) 
				char	fileName[65];
				#else
				char	fileName[1024];
				#endif
				char	sec[33];
				
				strcpy(sec, bc_sec);
				strcpy(fileName, bc_fileName);
				if	( strchr(code_line, '.') == NULL )
					strcat(code_line, ".bas");
				bc_load(code_line);
				strcpy(bc_fileName, fileName);
				strcpy(bc_sec, sec);
				}

			*crp = lc;
			}
		if	( (strncmp("SUB ", p, 4) == 0) || (strncmp("FUNC ", p, 5) == 0) )	{
			char	*lpar = NULL;
			char	*crp  = NULL;

			if	( *p == 'S' )
				p += 4;
			else
				p += 5;

			// close string
			lpar = strchr(p, '(');
			crp  = strchr(p, '\n');
			if	( lpar && crp )	{
				if	( lpar < crp )		{
					*lpar = '\0';
					crp = NULL;
					}
				else	{
					*crp = '\0';
					lpar = NULL;
					}
				}
			else if ( lpar )
				*lpar = '\0';
			else if ( crp )
				*crp = '\0';

			// add declaration
			bc_prepname(pname, p, 32);
			bc_addproc(pname);
			if	( bc_proc_level )	{
				strcat(bc_proc, "/");
				strcat(bc_proc, pname);
				}
			else
				strcpy(bc_proc, pname);

			bc_proc_level ++;

			// restore string
			if	( lpar )
				*lpar = '(';
			else	{
				if	( crp )
					*crp = '\n';
				}
			}
		else if ( bc_proc_level )	{
			if	( strncmp("END ", p, 4) == 0 || strncmp("END\n", p, 4) == 0 )	{
				char	*dol;

				dol = strrchr(bc_proc, '/');
				if	( dol )
					*dol = '\0';
				else
					*bc_proc = '\0';

				bc_proc_level --;
				}
			}

		// skip text line
		while ( *p != '\0' && *p != '\n' )
			p ++;

		if	( *p )
			p ++;
		}

	if	( bc_proc_level )
		sc_raise("FILE %s: SUB/FUNC END, MISSING", bc_fileName);

	bc_proc_level = 0;
	*bc_proc = '\0';

	dev_printf("File: \033\[1m%s\033\[0m\nSection: \033\[1m%s\033\[0m\n", bc_fileName, bc_sec);

	/*
	*	start
	*/
	if	( !scan_error )	{
		scan_line = 0;	// remove this for common line nums
		dev_printf("PASS1: Line %d", scan_line+1);
		ps = p = new_text;
		while ( *p )	{
			if	( *p == '\n' )	{
				*p = '\0';

				scan_line ++;
				if	( (scan_line % 32) == 0 )
					dev_printf("\rPASS1: Line %d", scan_line);
				bc_add2i(&bc_prog, kwTYPE_LINE, scan_line);

				strcpy(code_line, ps);
				bc_scan_cmd(code_line);
				if	( scan_error )
					break;

				ps = p+1;
				}

			if	( scan_error )
				break;

			p ++;
			}
		}

	tmp_free(code_line);
	tmp_free(new_text);

	bc_eoc(&bc_prog);
	bc_resize(&bc_prog, bc_prog.count);
	if	( !scan_error )	{
		dev_printf("\rPASS1: Line %d; finished\n", scan_line+1);
		#if !defined(_PalmOS)
		dev_printf("\rMemUsed: %d\n", memmgr_getalloc());
		#endif
		dev_printf("\n");
		}

}

/*
*	PASS 2
*/
void	bc_pass2(void) SEC(BCSCAN);
void	bc_pass2()
{
	dev_printf("PASS2...");
	if	( bc_proc_level )	
		sc_raise("MISSING SUB/FUNC-END");
	else if	( block_level )	
		sc_raise("MISSING LOOP-END");
	else	{
		first_data_ip = bc_prog.count;

		bc_pass2_scan();
		}

	if	( bc_data.count )	
		bc_append(&bc_prog, &bc_data);
}

/*
*	create bytecode
*/
mem_t	bc_createbin(void) SEC(BCSCAN);
mem_t	bc_createbin()
{
	word   		len, i;
	mem_t		buff_h;
	byte		*buff;
	label_t		label;

	dev_printf("Creating byte-code...\n");
	// total length
	len = 10; 					// num of vars + num of labels + def.data_ip + size of bytecode var
	len += (lab_count * 2);		// size of label table
	len += bc_prog.count;		// size of bytecode

	buff_h = mem_alloc(len+4); 	// +4 (checkop1233)
	buff = mem_lock(buff_h);

	// number of variables
	memcpy(buff, &var_count, 2);

	// number of labels
	memcpy(buff+2, &lab_count, 2);

	// label table
	for ( i = 0; i < lab_count; i ++ )	{
		dbt_read(lab_table, i, &label, sizeof(label_t));
		memcpy(buff+(i<<1)+4, &label.ip, 2);
		}

	// first data ip
	memcpy(buff+(lab_count<<1)+4, &first_data_ip, 2);

	// program length
	memcpy(buff+(lab_count<<1)+6, &bc_prog.count, 2);

	// the program itself
	memcpy(buff+(lab_count<<1)+8, bc_prog.ptr, bc_prog.count);

	dev_printf("Variables  %d\n", var_count);
	dev_printf("Labels     %d\n", lab_count);
	dev_printf("Proc/Func  %d\n", proc_count);
	dev_printf("Code size  %d\n", bc_prog.count);
	dev_printf("\n");
	mem_unlock(buff_h);
	return buff_h;	
}


