/*
*	strings!
*
*	Nicholas Christopoulos, 5/5/2000
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

#include "sys.h"
#include "str.h"
#include <stdarg.h>
#include "panic.h"

/*
*/
void	str_alltrim(char *str)
{
	char	*buf = tmp_alloc(strlen(str)+1);

	if	( *str == '\0' )	{
		tmp_free(buf);
		return;
		}
		
	if	( buf )	{
		char	*p = buf;

		strcpy(buf, str);
		while ( *p == ' ' || *p == '\t' )
			p ++;

		strcpy(str, p);
		p = str;
		while ( *p )	p ++;
		p--;
		while ( p >= str && (*p == ' ' || *p == '\t') )	p --;
		p++;
		*p = '\0';
			
		tmp_free(buf);
		}
}

/**
*/
char	*str_istr(const char *s1, const char *s2)
{
	char	*p;
	int		l2;

	p = (char *) s1;
	l2 = strlen(s2);
	while ( *p )	{
		if	( StrNCaselessCompare(p, s2, l2) == 0 )
			return p;
		p ++;
		}
	return NULL;
}

/*
*/
int		is_alpha(int ch)
{
	if	( ch == 0 )		return 0;
	if	( (ch > 64 && ch < 91) || (ch > 96 && ch < 123) )	return -1;
//	return (strchr("_", ch) != NULL); // Greek
	return (ch & 0x80);	// UTF8
}

/*
*	returns true if the ch is alphanumeric
*/
int		is_alnum(int ch)
{ 
	if	( ch == 0 )		return 0;
	return (is_alpha(ch) || is_digit(ch)); 
}							

/*
*	returns true if the ch is an 'empty' character
*/
int		is_space(int ch)
{ 
	return (ch == ' ' || ch == '\t' || ch == '\f' || ch == '\n' || ch == '\r' || ch == '\v') ? -1: 0; 
}

/*
*	returns true if 'text' contains digits only
*/
int		is_all_digits(const char *text)
{
	char	*p = (char *) text;

	if	( p == NULL )
		return 0;

	while ( *p )	{
		if	( !is_digit(*p) )
			return 0;
		p ++;
		}
	return 1;
}

/*
*	returns true if the text is keyword
*/
int		is_keyword(const char *name)
{
	char	*p = (char *) name;

	if	( p == NULL )
		return 0;

	if	( is_alpha(name[0]) )	{
		while ( is_alnum(*p) )	p ++;
		return (*p == '\0');
		}
	return 0;
}

/*
*	converts the 'str' string to uppercase
*/
#if !defined(_Win32)
char*	strupr(char *str)
{
	char	*p = str;

	if	( p == NULL )
		return 0;

	while ( *p )	{
		*p = to_upper(*p);
		p ++;
		}
	return str;
}
#endif

/*
*/
char	*get_keyword(char *text, char *dest)
{
	char	*p = (char *) text;
	char	*d = dest;

	if	( p == NULL )	{
		*dest = '\0';
		return 0;
		}

	while ( is_space(*p) )	p ++;

	while ( is_alnum(*p) )	{
		*d = to_upper(*p);
		d ++; p ++;
		}

//	Code to kill the $
//	if	( *p == '$' )	
//		p ++;

	if	( *p == '$' )
		*d ++ = *p ++;

	*d = '\0';
	while ( is_space(*p) )	p ++;
	return p;
}

/*
*	type  <=0 = error
*		    1 = int32
*           2 = double
*
*	Warning: octals are different from C (QB compatibility: 009 = 9)
*/
char	*get_numexpr(char *text, char *dest, int *type, long *lv, double *dv)
{
	char	*p = (char *) text;
	char	*d = dest;
	char	*epos = NULL;
	byte	base = 10;
	byte	dpc = 0;
	int		sign = 1;
	double	power = 1.0;

	*type = 0;
	*lv = 0;
	*dv = 0.0;

	if	( p == NULL )	{
		*dest = '\0';
		return NULL;
		}

	// spaces
	while ( is_space(*p) )	p ++;

	// sign
	if  ((*p == '-' || *p == '+') && strchr("0123356789.", *(p+1)) && *(p+1) != '\0')	{ 
		if	( *p == '-' )
			sign = -1;
		
		p ++;	// don't copy it
		}

	// hex|octal check
	if	( (*p == '&')
		  || 
		  (*p == '0' && (*(p+1) != '\0' && strchr("XBO", to_upper(*(p+1))) != NULL)
		  ))	{

		p ++;
		switch ( *p )	{
		case	'H':	case 'h':
		case	'X':	case 'x':
			base = 16;
			break;
		case	'O':	case 'o':
			base = 8;
			break;
		case	'B':	case 'b':
			base = 2;
			break;
		default:
			*type = -1;
			return p;	//	Unknown base
			}

		p ++;
		}

	// copy 
	if	( base == 16 )	{
		// copy hex
		while ( is_hexdigit(*p) )	{
			*d = to_upper(*p);
			d ++; p ++;
			}
		}
	else if	( base != 10 )	{
		// copy octal | bin
		while ( is_digit(*p) )	
			*d ++ = *p ++;
		}
	else {
		// copy number (first part)
		while ( is_digit(*p) || *p == '.' )	{
			if	( *p == '.' )	{
				dpc ++;
				if	( dpc > 1 )	{
					*type = -2;	// DP ERROR
					break;
					}
				}

			*d ++ = *p ++;
			}

		// check second part
		if ( (*p == 'E' || *p == 'e') && (*type == 0) )	{
			epos = d;
			*d ++ = *p ++;			// E

			if	( *p == '+' || *p == '-' || is_digit(*p) || *p == '.' )	{
				dpc = 0;
				
				// copy second part (power)
				if	( *p == '+' || *p == '-' )
					*d ++ = *p ++;
				while ( is_digit(*p) || *p == '.' )	{
					if	( *p == '.' )	{
						dpc ++;
						if	( dpc > 1 )	{
							*type = -4;	// DP ERROR (second part)
							break;
							}
						}
	
					*d ++ = *p ++;
					}
				}
			else	{
				*type = -3;	// E+- ERROR
				}
			}
		}

	*d = '\0';

	// calc. number
	if ( *type == 0 )	{
		switch ( base )	{
		case	10:
			if	( dpc || (epos != NULL) || (strlen(dest) > 8) )		{
				*type = 2;	// double
				if	( epos )	{
					*epos = '\0';
					power = pow(10, strtof(epos+1));
					*dv = strtof(dest) * ((double) sign) * power;
					*epos = 'E';
					}
				else	{
					*dv = strtof(dest) * ((double) sign);
					}
				}
			else	{
				*type = 1;	// int32
				*lv = xstrtol(dest) * sign;
				}
			break;
		case	16:
			*type = 1;	// int32
			*lv = hextol(dest);
			break;
		case	8:
			*type = 1;	// int32
			*lv = octtol(dest);
			break;
		case	2:
			*type = 1;	// int32
			*lv = bintol(dest);
			break;
			}
		}

	//
	if	( is_alpha(*p) )
		*type = -9;	// ITS NOT A NUMBER
	while ( is_space(*p) )	p ++;
	return p;
}

/*
*/
double	numexpr_strtof(char *source)
{
	char	buf[64], *np;
	int		type;
	long	lv;
	double	dv;

	np = get_numexpr(source, buf, &type, &lv, &dv);

	if	( type == 1 && *np == '\0' )	
		return (double) lv;
	else if	( type == 2 && *np == '\0' )	
		return dv;
	return 0.0;
}

/*
*/
long	numexpr_strtol(char *source)
{
	char	buf[64], *np;
	int		type;
	long	lv;
	double	dv;

	np = get_numexpr(source, buf, &type, &lv, &dv);

	if	( type == 1 && *np == '\0' )	
		return lv;
	else if	( type == 2 && *np == '\0' )	
		return (long) dv;
	return 0;
}

/*
*	convertion: binary to decimal
*/
long	bintol(const char *str)
{
	long	r = 0;
	char	*p = (char *) str;

	if	( p == NULL )	
		return 0;

	while ( *p )	{
		if	( *p == 48 || *p == 49 )	// 01
			r = (r << 1) + ((*p) - 48);
		p ++;
		}
	return r;
}

/*
*	convertion: octal to decimal
*/
long	octtol(const char *str)
{
	long	r = 0;
	char	*p = (char *) str;

	if	( p == NULL )	
		return 0;

	while ( *p )	{
		if	( *p >= 48 && *p <= 55 )	// 01234567
			r = (r << 3) + ((*p) - 48);
		p ++;
		}
	return r;
}

/*
*	convertion: hexadecimal to decimal
*/
long	hextol(const char *str)
{
	long	r = 0;
	char	*p = (char *) str;

	if	( p == NULL )	
		return 0;

	while ( *p )	{
		if	( is_digit(*p) )						// 0123456789
			r = (r << 4) + ((*p) - 48);
		else if	( *p >= 65 && *p <= 70 )		// ABCDEF
			r = (r << 4) + ((*p) - 55);
		else if	( *p >= 97 && *p <= 102 )		// abcdef
			r = (r << 4) + ((*p) - 87);
		p ++;
		}
	return r;
}

/*
*/
double	strtof(const char *str)
{
	char	*p = (char *) str;
	double	r = 0.0, d = 10.0;
	int		decp = 0;
	double	sign = 1;

	if	( p == NULL )	
		return 0;

	if	( *p == '-' )	{
		sign = -1;
		p ++;
		}
	else if	( *p == '+' )	
		p ++;
	
	while ( *p )	{
		if	( is_digit(*p) )	{
			if	( !decp )
				r = (r * 10) + ((*p) - 48);
			else	{
				r += (((*p) - 48) * 1 / d);
				d *= 10;
				}
			}
		else if ( *p == '.' )
			decp = 1;
		else if ( *p == ' ' )
			break;
		else	{
			r = 0;
			break;
			}

		p ++;
		}

	return r * ((double) sign);
}

/*
*/
long	xstrtol(const char *str)
{
	if	( str == NULL )	
		return 0;

	#if defined(_PalmOS)
	return StrAToI(str);
	#else
	return atoi(str);
	#endif
}

/*
*/
int		is_number(const char *str)
{
	char	*p = (char *) str;
	int		dpc = 0;

	if	( str == NULL )	
		return 0;

	if	( *p == '+' || *p == '-' )
		p ++;

	while ( *p )	{
		if	( strchr("0123456789.", *p) == NULL )
			return 0;
		if	( *p == '.' )	{
			dpc ++;
			if	( dpc > 1 )
				return 0;
			}
		p ++;
		}
	return 1;
}

/*
*/
char	*ftostr(double num, char *dest)
{
	#if defined(_PalmOS)
	long	i, l, dc;
	double	f = num;
	char	s[64], *p, *lz;

	if	( dest == NULL )	
		panic("f2s(..,null)");

	// zero
	*dest = '\0';
	if	( f == 0.0 )	{
		strcpy(dest, "0");
		return dest;	
		}
	
	// negative
	if	( f < 0.0 )	
		f = f * -1.0;

	// find the power
	dc = 0;
	if	( f >= 1.0 )	{
		while ( f >= 1.0 )	{
			f = f / 10.0;
			dc ++;
			}
		}
	else if ( f < 0.1 )	{
		while ( f < 0.1 )	{
			f = f * 10;
			dc --;
			}
		}

	// 
	l = (dword) (round(f * 1000000000.0));
	StrIToA(s, l);

	if	( dc <= 6 && dc > 0 )	{	// 1.0 =< abs(n) < x*10^6 
		strncpy(dest, s, dc);
		dest[dc] = '.';
		dest[dc+1] = '\0';
		strcat(dest, (s+dc));

		// remove zeroes
		p = dest;
		lz = NULL;
		while ( *p )	{
			if	( *p == '.' )
				break;
			p ++;
			}

		if	( *p )	{
			while ( *p )	{
				if	( *p != '0' )
					lz = p;
				p ++;
				}
			if	( lz )	{
				if	( *lz == '.' )
					*lz = '\0';
				else
					*(lz+1) = '\0';
				}
			}

		// sign
		if	( num < 0.0 )	{
			strcpy(s, dest);
			strcpy(dest, "-");
			strcat(dest, s);
			}
		}
	else	{						// 1.0 > abs(n) >= 0.0
		if	( num < 0.0 )
			strcpy(dest, "-0.");
		else
			strcpy(dest, "0.");

		if	( dc == -1 || dc == -2 || dc == -3 )	{
			for ( i = 0; i < ABS(dc); i ++ )	
				strcat(dest, "0");
			}

		strcat(dest, s);

		// remove zeroes
		p = dest;
		lz = NULL;
		while ( *p )	{
			if	( *p != '0' )
				lz = p;
			p ++;
			}
		if	( lz )
			*(lz+1) = '\0';

		// power
		if	( dc != 0 && dc != -1 && dc != -2 && dc != -3  )	{
			if	( dc >= 0 )
				strcat(dest, "E+");
			else
				strcat(dest, "E-");
			StrIToA(s, dc);
			strcat(dest, s);
			}
		}
	#else
	sprintf(dest, "%g", num);
	#endif
	return dest;
}

/*
*/
char	*ltostr(long num, char *dest)
{
	if	( dest == NULL )	
		panic("l2s(..,null)");

	#if defined(_PalmOS)
	StrIToA(dest, num);
	#else
	sprintf(dest, "%ld", num);
	#endif
	return dest;
}

#if defined(_PalmOS)
char	*strrchr(const char *source, int ch)
{
	char	*p = (char *) source;
	char	*last_p = NULL;

	while ( *p )	{
		if	( *p == ch )
			last_p = p;
		p ++;
		}

	return last_p;
}
#endif

